﻿using System;

namespace AlgorithmByCSharp.BubbleSort
{
    class BubbleSortBorder
    {
        /// <summary>
        /// 冒泡排序
        /// 思想：相邻元素两两比较，当一个元素大于右侧相邻元素时，交换它们的位置；当一个元素小于或等于右侧相邻元素时，位置不变。
        /// 上述代码中，sortBorder就是无序数列的边界。
        /// 在每一轮排序过程中，处于sortBorder之后的元素就不需要进行比较排序了，肯定是有序的。
        /// </summary>
        public void BubbleSort(int[] array)
        {
            //记录最后一次交换的位置
            int lastExchangeIndex = 0;
            //无序数列的边界，每次比较只需要比到这里为止
            int sortBorder = array.Length - 1;  // sortBorder 有序区标记
            for (int i = 0; i < array.Length - 1; i++)
            {
                //有序标记，每一轮的初始值都是true
                bool isSorted = true;
                for (int j = 0; j < sortBorder; j++)
                {
                    int temp = 0;
                    if (array[j] > array[j + 1])
                    {
                        temp = array[j];
                        array[j] = array[j + 1];
                        array[j + 1] = temp;
                        //因为有元素进行交换，所以不是有序的，标记为false
                        isSorted = false;
                        //更新为最后一次交换元素的位置
                        lastExchangeIndex = j;
                    }
                }
                sortBorder = lastExchangeIndex;
                if (isSorted) { break; }
            }
        }

        /// <summary>
        /// 鸡尾酒排序
        /// 代码外层的大循环控制着所有排序回合，大循环内包含2个小循环，
        /// 第1个小循环从左向右比较并交换元素，
        /// 第2个小循环从右向左比较并交换元素
        /// 适用场合：大部分元素已经有序的情况下
        /// </summary>
        public void ShakerSort(int[] array)
        {
            int temp = 0;
            for (int i = 0; i < array.Length / 2; i++)
            {
                // 有序标记，每一轮的初始值都是true
                bool isSorted = true;
                // 奇数轮，从左到右比较和交换
                for (int j = i; j < array.Length - i - 1; j++)
                {
                    if (array[j] > array[j + 1])
                    {
                        temp = array[j];
                        array[j] = array[j + 1];
                        array[j + 1] = temp;
                        // 有元素交换，所以不是有序的，标记变为false
                        isSorted = false;
                    }
                }
                if (isSorted) break;
                // 在偶数轮之前，将isSorted重新标记为true
                isSorted = true;
                for (int j = array.Length - i - 1; j > i; j--)
                {
                    if (array[j] < array[j - 1])
                    {
                        temp = array[j];
                        array[j] = array[j - 1];
                        array[j - 1] = temp;
                        // 因为有元素进行交换，所以不是有序的，标记变为false
                        isSorted = false;
                    }
                }
                if (isSorted) { break; }
            }
        }
    }
}
